/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef BITINTEGER_H
#define BITINTEGER_H

#include <typeinfo>

#include <bit/enums.h>
#include <bit/utility.h>
#include <bit/pointer.h>
#include <bit/data.h>

namespace bit
{

  template <typename I> class Int;

  class Integer
  {
    public:

      typedef BitPointer<Integer> pointer;

      Integer( ByteOrder bo=BYTEORDER_HOST ): byte_order(bo) { }

      virtual ~Integer() { }

      virtual Data as_data() const = 0;

      virtual size_t size() const = 0;

      virtual const void* voidptr() const = 0;

      virtual pointer clone() const = 0;

      /**
       * Sets the value from a data object by directly copying; byte ordering
       * of the data is not considered.
       * @param d The data object. The first n bytes are used to set the value.
       */
      virtual void set_value( const Data d ) = 0;

      virtual void set_value( const void* mem, size_t size ) = 0;

      /**
       * Sets the value from a data object
       * @param d The data object. The first n bytes are used to set the value.
       * @param databo The byte order of the data object
       */
      virtual void set_value( const Data d, ByteOrder databo ) = 0;

      template <typename T> operator T();

      template <typename T> Integer& operator = ( T i );

      virtual const std::type_info& int_type() = 0;

      ByteOrder byte_order;
  };

  template <typename I>
  class Int: public Integer {
    public:

      typedef BitPointer< Int<I> > pointer;

      Int(I v=0, ByteOrder bo=BYTEORDER_HOST): Integer(bo), value(v) { }

      Int(ByteOrder bo): Integer(bo), value(0) { }

      /**
       * Constructor that sets the value from a data object
       * @param d The data object. The first n bytes are used to set the value.
       * @param databo The byte order of the data object
       * @param bo The byte order of this integer.
       */
      Int(const Data d, ByteOrder databo=BYTEORDER_HOST, ByteOrder bo=BYTEORDER_HOST): Integer(bo) {
        this->set_value(d, databo);
      }

      static pointer create(I v=0, ByteOrder bo=BYTEORDER_HOST) { return pointer(new Int<I>(v,bo)); }

      static pointer create(ByteOrder bo) { return pointer(new Int<I>(bo)); }

      static pointer create(const Data d, ByteOrder databo=BYTEORDER_HOST, ByteOrder bo=BYTEORDER_HOST) {
        return pointer(new Int<I>(d,databo,bo));
      }

      ~Int() { }

      virtual Integer::pointer clone() const {
        pointer p = this->create(value, byte_order);
        return p;
      }

      operator I() const { return this->host(); }

      bool operator <  ( I i ) const { return this->host() <  i; }
      bool operator <= ( I i ) const { return this->host() <= i; }
      bool operator == ( I i ) const { return this->host() == i; }
      bool operator != ( I i ) const { return this->host() != i; }
      bool operator >= ( I i ) const { return this->host() >= i; }
      bool operator >  ( I i ) const { return this->host() >  i; }

      Int<I>& operator = ( I i )
      {
        value = i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I> operator + ( I i ) const
      {
        this->convert_value_from_type_to_host();
        value += i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I> operator - ( I i ) const
      {
        this->convert_value_from_type_to_host();
        value -= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I> operator * ( I i ) const
      {
        this->convert_value_from_type_to_host();
        value *= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I> operator / ( I i ) const
      {
        this->convert_value_from_type_to_host();
        value /= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I> operator % ( I i ) const
      {
        this->convert_value_from_type_to_host();
        value %= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I>& operator += ( I i )
      {
        this->convert_value_from_type_to_host();
        value += i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I>& operator -= ( I i )
      {
        this->convert_value_from_type_to_host();
        value -= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I>& operator *= ( I i )
      {
        this->convert_value_from_type_to_host();
        value *= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I>& operator /= ( I i )
      {
        this->convert_value_from_type_to_host();
        value /= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Int<I>& operator %= ( I i )
      {
        this->convert_value_from_type_to_host();
        value %= i;
        this->convert_value_from_host_to_type();
        return *this;
      }

      /**
       * Sets the value from a data object by directly copying; byte ordering
       * of the data is not considered.
       * @param d The data object. The first n bytes are used to set the value.
       */
      virtual void set_value( const Data d ) {
        if ( d.size() < sizeof(I) )
          throw;

        memcpy( &value, d.data(), sizeof(I) );
      }

      /**
       * Sets the value from a data object by directly copying; byte ordering
       * of the data is not considered.
       */
      virtual void set_value( const void* mem, size_t size ) {
        if ( size < sizeof(I) )
          throw;

        memcpy( &value, mem, sizeof(I) );
      }

      /**
       * Sets the value from a data object
       * @param d The data object. The first n bytes are used to set the value.
       * @param databo The byte order of the data object
       */
      virtual void set_value( const Data d, ByteOrder databo ) {
        if ( d.size() < sizeof(I) )
          throw;

        memcpy( &value, d.data(), sizeof(I) );

        switch ( byte_order ) {
          case BYTEORDER_HOST:
            switch ( databo ) {
              case BYTEORDER_HOST:
                break;
              case BYTEORDER_NETWORK:
              case BYTEORDER_BIG_ENDIAN:
                value = be_to_host(value);
                break;
              case BYTEORDER_LITTLE_ENDIAN:
                value = le_to_host(value);
                break;
            }
            break;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            switch ( databo ) {
              case BYTEORDER_HOST:
                value = host_to_net(value);
                break;
              case BYTEORDER_NETWORK:
              case BYTEORDER_BIG_ENDIAN:
                break;
              case BYTEORDER_LITTLE_ENDIAN:
                value = le_to_net(value);
                break;
            }
            break;
          case BYTEORDER_LITTLE_ENDIAN:
            switch ( databo ) {
              case BYTEORDER_HOST:
                value = host_to_le(value);
                break;
              case BYTEORDER_NETWORK:
              case BYTEORDER_BIG_ENDIAN:
                value = net_to_le(value);
                break;
              case BYTEORDER_LITTLE_ENDIAN:
                break;
            }
            break;
        }
      }

      virtual Data as_data() const
      {
        Data d( sizeof(I) );
        memcpy( d.data(), &value, sizeof(I) );
        return d;
      }

      virtual size_t size() const { return sizeof(I); }

      virtual const void* voidptr() const { return &value; }

      virtual const std::type_info& int_type() { return typeid(I); }

      I host() const {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return value;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            return net_to_host(value);
          case BYTEORDER_LITTLE_ENDIAN:
            return le_to_host(value);
        }
        throw;
      }

      I network() const {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return host_to_net(value);
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            return value;
          case BYTEORDER_LITTLE_ENDIAN:
            return le_to_net(value);
        }
        throw;
      }

      I big_endian() const {
        return this->network();
      }

      I little_endian() const {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return host_to_le(value);
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            return net_to_le(value);
          case BYTEORDER_LITTLE_ENDIAN:
            return value;
        }
        throw;
      }

      I value;

    protected:
      void convert_value_from_type_to_host()
      {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            value = net_to_host(value);
          case BYTEORDER_LITTLE_ENDIAN:
            value = le_to_host(value);
        }
      }

      void convert_value_from_host_to_type()
      {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            value = host_to_net(value);
          case BYTEORDER_LITTLE_ENDIAN:
            value = host_to_le(value);
        }
      }

  };

  typedef Int<uint8_t> UInt8;
  typedef Int<int8_t> Int8;
  typedef Int<uint16_t> UInt16;
  typedef Int<int16_t> Int16;
  typedef Int<uint32_t> UInt32;
  typedef Int<int32_t> Int32;
  typedef Int<uint64_t> UInt64;
  typedef Int<int64_t> Int64;

  template <typename T> Integer::operator T()
  {
    switch ( this->size() )
    {
      case 1:
        if ( this->int_type() == typeid(uint8_t) )
        {
          UInt8* uint8;
          uint8 = dynamic_cast<UInt8*>(this);
          return static_cast<T>(uint8->host());
        }
        else if ( this->int_type() == typeid(int8_t) )
        {
          Int8* int8;
          int8 = dynamic_cast<Int8*>(this);
          return static_cast<T>(int8->host());
        }
        break;
      case 2:
        if ( this->int_type() == typeid(uint16_t) )
        {
          UInt16* uint16;
          uint16 = dynamic_cast<UInt16*>(this);
          return static_cast<T>(uint16->host());
        }
        else if ( this->int_type() == typeid(int16_t) )
        {
          Int16* int16;
          int16 = dynamic_cast<Int16*>(this);
          return static_cast<T>(int16->host());
        }
        break;
      case 4:
        if ( this->int_type() == typeid(uint32_t) )
        {
          UInt32* uint32;
          uint32 = dynamic_cast<UInt32*>(this);
          return static_cast<T>(uint32->host());
        }
        else if ( this->int_type() == typeid(int32_t) )
        {
          Int32* int32;
          int32 = dynamic_cast<Int32*>(this);
          return static_cast<T>(int32->host());
        }
        break;
      case 8:
        if ( this->int_type() == typeid(uint64_t) )
        {
          UInt64* uint64;
          uint64 = dynamic_cast<UInt64*>(this);
          return static_cast<T>(uint64->host());
        }
        else if ( this->int_type() == typeid(int64_t) )
        {
          Int64* int64;
          int64 = dynamic_cast<Int64*>(this);
          return static_cast<T>(int64->host());
        }
    }
    throw std::logic_error("bit::Integer bad cast");
  }

  template <typename T> Integer& Integer::operator = ( T i )
  {
    switch ( this->size() )
    {
      case 1:
        if ( this->int_type() == typeid(uint8_t) )
        {
          UInt8* uint8;
          uint8 = dynamic_cast<UInt8*>(this);
          *uint8 = (uint8_t)(i);
        }
        else if ( this->int_type() == typeid(int8_t) )
        {
          Int8* int8;
          int8 = dynamic_cast<Int8*>(this);
          *int8 = (int8_t)(i);
        }
        break;
      case 2:
        if ( this->int_type() == typeid(uint16_t) )
        {
          UInt16* uint16;
          uint16 = dynamic_cast<UInt16*>(this);
          *uint16 = (uint16_t)(i);
        }
        else if ( this->int_type() == typeid(int16_t) )
        {
          Int16* int16;
          int16 = dynamic_cast<Int16*>(this);
          *int16 = (int16_t)(i);
        }
        break;
      case 4:
        if ( this->int_type() == typeid(uint32_t) )
        {
          UInt32* uint32;
          uint32 = dynamic_cast<UInt32*>(this);
          *uint32 = (uint32_t)(i);
        }
        else if ( this->int_type() == typeid(int32_t) )
        {
          Int32* int32;
          int32 = dynamic_cast<Int32*>(this);
          *int32 = (int32_t)(i);
        }
        break;
      case 8:
        if ( this->int_type() == typeid(uint64_t) )
        {
          UInt64* uint64;
          uint64 = dynamic_cast<UInt64*>(this);
          *uint64 = (uint64_t)(i);
        }
        else if ( this->int_type() == typeid(int64_t) )
        {
          Int64* int64;
          int64 = dynamic_cast<Int64*>(this);
          *int64 = (int64_t)(i);
        }
    }
    return *this;
  }

}

#endif
