/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef BITRECORDVECTOR_H
#define BITRECORDVECTOR_H

#include <vector>

#include <bit/recordbase.h>
#include <bit/recordstorage.h>

namespace bit {

/**
 * Implements a vector of records, synchronizing all fields within each record.
 *
 * @todo Enforce synchronization among Fields; probably means that signals are needed
 *
 * @ingroup records
 * @author Rick L. Vinyard, Jr. <rvinyard@cs.nmsu.edu>
 */
class RecordVector : public RecordBase
{
public:

  class Element: public FieldBase
  {
    public:

    typedef BitPointer<Element> pointer;

    protected:
      friend class RecordVector;

      Element( size_t index, RecordStorage& reference_layout );

      static Element::pointer create( size_t index, RecordStorage& reference_layout );

    public:

      /** Destructor */
      ~Element();

      virtual size_t length() const;

      virtual size_t length(size_t units) const;

      virtual size_t length_units() const;

      virtual int offset() const;

      virtual int offset(size_t units) const;

      virtual size_t offset_units() const;

      virtual std::string name() const;

      virtual std::string name(int depth) const;

      virtual std::string description() const;

      virtual const FieldType& type() const;

      virtual void set_type( const FieldType& );

      /** Returns the data encoding of the parent container */
      virtual Encoding encoding() const;

      /** Sets the data encoding of the parent container */
      virtual void set_encoding( Encoding en );

      // TODO do something real with this
      virtual const std::string& xml() { return m_xml; }

      virtual iterator begin();

      virtual iterator end();

      virtual size_t subfields();

      virtual FieldBase::pointer subfield(const std::string& s);

      virtual FieldBase::pointer subfield(size_t i);

      virtual FieldBase::pointer clone();

      size_t index();

      virtual Container container_type() { return RECORD_ELEMENT; }

    protected:
      size_t m_index;

      RecordStorage m_fields;

      virtual FieldBase::pointer previous_field(FieldBase::pointer current_field) throw (exception::invalid_container_op);

      virtual FieldBase::pointer next_field(FieldBase::pointer current_field) throw (exception::invalid_container_op);
  };

  /** Convenience typedef to a smart pointer to bit::Record */
  typedef BitPointer<RecordVector> pointer;

  /** Default constructor */
  RecordVector( std::string name="",
          size_t size=0,
          size_t length_units=BITS,
          int offset=0,
          size_t offset_units=BITS,
          std::string descr=""
        );

  /** Default constructor */
  RecordVector( const RecordStorage& layout,
                std::string name="",
                size_t size=0,
                size_t length_units=BITS,
                int offset=0,
                size_t offset_units=BITS,
                std::string descr=""
              );

  /** Static method returns a smart pointer to a new Record object */
  static RecordVector::pointer create( std::string name="",
                                 size_t size=0,
                                 size_t length_units=BITS,
                                 int offset=0,
                                 size_t offset_units=BITS,
                                 std::string descr=""
                               );

  /** Static method returns a smart pointer to a new Record object */
  static RecordVector::pointer create( const RecordStorage& layout,
                                       std::string name="",
                                       size_t size=0,
                                       size_t length_units=BITS,
                                       int offset=0,
                                       size_t offset_units=BITS,
                                       std::string descr=""
                                     );

  /** Destructor */
  ~RecordVector();

  /** @return the length of this object in it's length units */
  virtual size_t length() const;

  virtual size_t length(size_t units) const;

  size_t size() const;

  void set_size(size_t size);

  virtual const std::string& xml();

  virtual iterator begin();

  virtual iterator end();

  virtual size_t subfields();

  virtual FieldBase::pointer subfield(const std::string& s);

  virtual FieldBase::pointer subfield(size_t i);

  virtual void add_field(FieldBase::pointer child);

  virtual void remove_field(const std::string& name);

  virtual FieldBase::pointer clone();

  virtual Container container_type() { return RECORD_VECTOR; }

  protected:
    RecordStorage m_record_layout;
    size_t m_size;

    typedef std::vector<Element::pointer> Elements;

    Elements m_elements;

    virtual FieldBase::pointer previous_child(FieldBase::pointer current_child) throw (exception::invalid_container_op);

    virtual FieldBase::pointer next_child(FieldBase::pointer current_child) throw (exception::invalid_container_op);

};

}

#endif
